from __future__ import annotations

from logging import getLogger
from typing import Any

from jira_select.plugin import BaseFunction

from ..exceptions import JiraSelectError
from ..exceptions import QueryError
from ..query import Executor

logger = getLogger(__name__)


class Function(BaseFunction):
    def __call__(  # type: ignore[override]
        self, subquery_name: str, **params: dict[str, Any]
    ) -> list[Any]:
        params = params if params is not None else {}
        results: list[Any] = []

        if not self.query:
            raise JiraSelectError("Parent query unexpectedly unavailable to subquery.")

        try:
            query_definition = self.query.subqueries[subquery_name]
        except KeyError as exc:
            raise QueryError(
                f"Subquery '{subquery_name}' does not exist in query definition."
            ) from exc

        # Unless explicitly set, subqueries use the same
        # caching duration as their parent query
        if query_definition.cache is None and self.query.cache:
            query_definition.cache = self.query.cache

        query = Executor(
            self.jira,
            query_definition,
            progress_bar=False,
            parameters=params,
            schema=self.executor.schema if self.executor else None,
        )
        for row in query:
            row_values = list(row.values())
            if len(query_definition.select) == 1:
                results.append(row_values[0])
            else:
                results.append(row_values)

        return results



from typing import Dict
from typing import Optional

from jira.resources import Sprint

from ..plugin import BaseFunction


class Function(BaseFunction):
    """Returns a `jira.resources.Sprint` object for a given sprint ID string."""

    CACHE: Dict[int, Sprint] = {}

    def get_sprint_details(self, id: int) -> Sprint:
        if id not in self.CACHE:
            sprint = self.jira.sprint(id)
            self.CACHE[id] = Sprint(
                {"agile_rest_path": self.jira._options["agile_rest_path"]},
                None,
                sprint.raw,
            )

        return self.CACHE[id]

    def __call__(self, id: Optional[int]) -> Optional[Sprint]:  # type: ignore[override]
        if id is None:
            return None

        return self.get_sprint_details(id)



import logging
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import List

from dotmap import DotMap
from jira import JIRA
from jira.resources import Issue
from simpleeval import NameNotDefined

from ..exceptions import ExpressionParameterMissing
from ..exceptions import QueryError
from ..plugin import BaseSource
from ..plugin import get_installed_functions
from ..types import SchemaRow
from ..utils import evaluate_expression
from ..utils import find_missing_parameters

logger = logging.getLogger(__name__)


class Source(BaseSource):
    SCHEMA: List[SchemaRow] = [
        SchemaRow.parse_obj({"id": "key", "type": "str"}),
        SchemaRow.parse_obj({"id": "id", "type": "int"}),
    ]

    @classmethod
    def get_field_data(
        cls, row: Any, expression: str, functions: Dict[str, Callable]
    ) -> str:
        result = evaluate_expression(expression, row, functions=functions)
        if isinstance(result, str):
            return result
        return ""

    @classmethod
    def get_schema(cls, jira: JIRA) -> List[SchemaRow]:
        field_definitions: List[SchemaRow] = super().get_schema(jira)
        functions = get_installed_functions()

        for column in jira.fields():
            try:
                type = str(cls.get_field_data(DotMap(column), "schema.type", functions))
            except NameNotDefined:
                type = ""
            field_definitions.append(
                SchemaRow.parse_obj(
                    {
                        "id": str(cls.get_field_data(DotMap(column), "id", functions)),
                        "type": type,
                        "description": str(
                            cls.get_field_data(DotMap(column), "name", functions)
                        ),
                        "raw": DotMap(column),
                    }
                )
            )

        return field_definitions

    def _get_jql(self) -> str:
        if self.query.where and not isinstance(self.query.where, list):
            raise QueryError(
                "Issue queries 'where' should be a list of JQL expression strings."
            )

        assert isinstance(self.query.where, list)

        query = " AND ".join(f"({q})" for q in self.query.where)

        if missing := find_missing_parameters(
            query, list(self._executor.parameters.keys())
        ):
            raise ExpressionParameterMissing(
                "Parameter {params.%s} found in expression, but no parameter was specified!"
                % missing[0]
            )

        query = query.format(params=DotMap(self._executor.parameters))

        order_by_fields = ", ".join(self.query.order_by)

        if order_by_fields:
            query = f"{query} ORDER BY {order_by_fields}"

        logger.debug("Executing JIRA query with JQL %s", query)

        return query

    def __iter__(self) -> Iterator[Dict]:
        start_at = 0
        max_results = 2**32
        result_limit = self.query.limit or 2**32

        jql = self._get_jql()

        self.update_progress(completed=0, total=1, visible=True)
        while start_at < min(max_results, result_limit):
            results = self.jira.search_issues(
                jql,
                startAt=start_at,
                expand=",".join(self.query.expand),
                fields="*all",
                maxResults=min(result_limit, 100),
            )

            max_results = results.total
            count = min([results.total, result_limit])
            self.update_count(count)

            for result in results:
                self.update_progress(advance=1, total=count, visible=True)

                yield result.raw

                start_at += 1

                # Return early if our result limit has been reached
                if start_at >= result_limit:
                    break

    def rehydrate(self, value: Dict) -> Issue:
        return Issue({}, None, value)



try:
    import debugpy
except ImportError:
    debugpy = None


def pytest_addoption(parser):
    parser.addoption("--debugger", action="store_true")
    parser.addoption("--debugger-port", type=int, default=5678)


def pytest_sessionstart(session):
    if not debugpy:
        return

    if session.config.getoption("debugger"):
        port = session.config.getoption("debugger_port")
        debugpy.listen(port)
        debugpy.wait_for_client()



from __future__ import annotations

from typing import Any

import portion
from jira.resources import Issue

from jira_select.plugin import BaseFunction

from .get_issue_snapshot_on_date import snapshot_iterator
from .simple_filter import simple_filter


class Function(BaseFunction):
    def __call__(  # type: ignore[override]
        self, issue: Issue, **filter_params: dict[str, Any]
    ) -> portion.Interval:
        interval = portion.empty()

        matching_snapshots = simple_filter(
            list(
                snapshot_iterator(
                    issue, self.executor.field_name_map if self.executor else {}
                )
            ),
            **filter_params,
        )

        for snapshot in matching_snapshots:
            interval |= portion.closed(snapshot.validity_start, snapshot.validity_end)

        return interval



import argparse
import os
import shutil
import subprocess
import tempfile
from typing import IO
from typing import Optional
from typing import Type

from jira import JIRAError
from prompt_toolkit import PromptSession
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.history import FileHistory
from prompt_toolkit.lexers import PygmentsLexer
from pygments.lexers.data import YamlLexer
from yaml import safe_load

from .. import __version__
from ..constants import DEFAULT_INLINE_VIEWERS
from ..exceptions import QueryError
from ..plugin import BaseCommand
from ..plugin import BaseFormatter
from ..plugin import get_installed_formatters
from ..plugin import get_installed_functions
from ..query import Executor
from ..types import QueryDefinition
from ..utils import get_config_dir
from ..utils import launch_default_viewer


class QueryParseError(Exception):
    pass


class Command(BaseCommand):
    @classmethod
    def get_help(cls) -> str:
        return (
            "Opens an interactive shell (a.k.a. repl) allowing you to "
            "interact with Jira and see results immediately (like "
            "the sqlite3, postgres, or mysql shells)."
        )

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--editor-mode",
            "-m",
            choices=["emacs", "vi"],
            default=None,
        )
        parser.add_argument(
            "--disable-progressbars",
            "-b",
            action="store_false",
            default=True,
            dest="enable_progressbars",
        )
        parser.add_argument(
            "--output",
            "-o",
            type=argparse.FileType("wb+"),
            default=None,
            help=(
                "Path to file where records will be written; default: "
                "a temporary file."
            ),
        )
        parser.add_argument(
            "--launch-default-viewer",
            "-l",
            action="store_true",
            default=False,
            help=(
                "Display the output using the default viewer for the "
                "filetype used by the selected formatter instead of "
                "displaying the results inline.  Depending upon your "
                "operating system, you may need to explicitly specify "
                "the output file path using the --output parameter "
                "for this option to work."
            ),
            dest="launch_default_viewer",
        )
        parser.add_argument(
            "--format",
            "-f",
            default="json",
            choices=get_installed_formatters().keys(),
            dest="format",
        )

    def _prompt_loop(
        self,
        session: PromptSession,
        formatter_cls: Type[BaseFormatter],
        outf: IO[bytes],
    ):
        result = session.prompt(">>> ")

        try:
            query_definition: QueryDefinition = QueryDefinition.parse_obj(
                safe_load(result)
            )
            query = Executor(
                self.jira,
                query_definition,
                progress_bar=self.options.enable_progressbars,
            )
        except Exception as e:
            raise QueryParseError(e)

        outf.seek(0)
        outf.truncate()

        with formatter_cls(query, outf) as formatter:
            for row in query:
                formatter.writerow(row)
                outf.flush()

        if self.options.launch_default_viewer:
            launch_default_viewer(outf.name)
        else:
            self._inline_viewer(outf)

    def _inline_viewer(self, fobj: IO[bytes]):
        viewer: Optional[str] = self.config.inline_viewers.get(
            self.options.format, DEFAULT_INLINE_VIEWERS.get(self.options.format)
        )

        if viewer and shutil.which(viewer):
            proc = subprocess.Popen([viewer, fobj.name])
            proc.wait()
        else:
            fobj.seek(0)
            self.console.print(fobj.read().decode("utf-8"))

    def build_completions(self) -> WordCompleter:
        sql_completions = [
            "select",
            "from",
            "where",
            "order_by",
            "having",
            "group_by",
            "sort_by",
            "expand",
            "limit",
            "cap",
        ]
        function_completions = list(get_installed_functions(self.jira).keys())
        field_completions = [field["id"] for field in self.jira.fields()]

        return WordCompleter(sql_completions + function_completions + field_completions)

    def handle(self) -> None:
        self.console.print(
            f"[bold]Jira-select[/bold] Shell v{__version__}",
            style="dodger_blue1 blink",
        )
        vi_mode = not self.config.shell.emacs_mode
        if self.options.editor_mode:
            vi_mode = self.options.editor_mode
        if vi_mode:
            self.console.print(
                " | [bold]Run:[/bold]\t\tESC->ENTER",
                style="deep_sky_blue4",
            )
            self.console.print(
                " | [bold]Clear:[/bold]\tCTRL+C",
                style="deep_sky_blue4",
            )
            self.console.print(
                " | [bold]Exit:[/bold]\tCTRL+D",
                style="deep_sky_blue4",
            )

        completions = self.build_completions()
        session: PromptSession = PromptSession(
            lexer=PygmentsLexer(YamlLexer),
            multiline=True,
            completer=completions,
            complete_while_typing=False,
            history=FileHistory(os.path.join(get_config_dir(), "shell_history")),
            auto_suggest=AutoSuggestFromHistory(),
            vi_mode=vi_mode,
            mouse_support=True,
        )
        formatter_cls = get_installed_formatters()[self.options.format]

        output = self.options.output
        if not output:
            output = tempfile.NamedTemporaryFile(
                "wb+",
                suffix=f".{formatter_cls.get_file_extension()}",
            )

        while True:
            try:
                self._prompt_loop(session, formatter_cls, output)
            except JIRAError as e:
                self.console.print(f"[red][bold]Jira Error:[/bold] {e.text}[/red]")
            except QueryError as e:
                self.console.print(f"[red][bold]Query Error:[/bold] {e}[/red]")
            except QueryParseError as e:
                self.console.print(
                    f"[red][bold]Parse Error:[/bold] Your query could not be parsed: {e}[/red]"
                )
            except KeyboardInterrupt:
                continue
            except EOFError:
                break
            except Exception:
                self.console.print_exception()

        output.close()



from io import TextIOWrapper
from typing import Any
from typing import Dict
from typing import List

from rich.console import Console
from rich.table import Table

from ..plugin import BaseFormatter


class Formatter(BaseFormatter):
    @classmethod
    def get_file_extension(cls) -> str:
        return "txt"

    def _generate_fieldnames(self) -> List[str]:
        fields = []

        for field in self.executor.query.select:
            fields.append(field.column)

        return fields

    def open(self):
        super().open()
        self.table = Table()

        for field in self.executor.query.select:
            self.table.add_column(field.column)

    def close(self):
        wrapped = TextIOWrapper(self.stream, encoding="utf-8", write_through=True)
        console = Console(file=wrapped)
        console.print(self.table)
        wrapped.detach()
        super().close()

    def writerow(self, row: Dict[str, Any]):
        fields = [str(row.get(field.column)) for field in self.executor.query.select]
        self.table.add_row(*fields)



import argparse
import json

from rich.table import Table

from ..plugin import BaseCommand
from ..utils import JiraSelectJsonEncoder


class Command(BaseCommand):
    @classmethod
    def get_help(cls) -> str:
        return (
            "Displays Jira instances that you have configured jira-select "
            "to be able to use."
        )

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--json",
            "-j",
            default=False,
            action="store_true",
            help="Export schema as JSON format instead of printing a table.",
        )

    def handle(self) -> None:
        if self.options.json:
            instances = []
            for instance_name, instance_data in self.config.instances.items():
                instances.append(
                    {
                        "name": instance_name,
                        "username": instance_data.username,
                        "url": instance_data.url,
                    }
                )
            self.console.print(
                json.dumps(
                    instances, cls=JiraSelectJsonEncoder, sort_keys=True, indent=4
                )
            )
        else:
            table = Table(title="Configured Instances")
            table.add_column("name", style="green")
            table.add_column("username", style="cyan")
            table.add_column("url", style="bright_cyan")

            for instance_name, instance_data in self.config.instances.items():
                table.add_row(instance_name, instance_data.username, instance_data.url)

            self.console.print(table)



import datetime
import re
from dataclasses import dataclass
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import cast

from dateutil.parser import parse as parse_date

from ..plugin import BaseFunction


@dataclass
class SprintInfo:
    id: int = -1
    name: str = ""
    state: str = ""
    goal: str = ""
    rapidViewId: Optional[int] = -1
    sequence: Optional[int] = -1
    startDate: Optional[datetime.datetime] = None
    endDate: Optional[datetime.datetime] = None
    completeDate: Optional[datetime.datetime] = None


class Function(BaseFunction):
    """Returns a `jira_select.functions.sprint_details.SprintInfo` object for a given sprint ID string."""

    def get_sprint_details(self, representation: str) -> Optional[SprintInfo]:
        # com.atlassian.greenhopper.service.sprint.Sprint@14b1c359[id=436,rapidViewId=153,state=CLOSED,name=ORC #11 3/9-3/23,goal=Code Complete 3/18,startDate=2020-03-09T21:53:07.264Z,endDate=2020-03-23T20:53:00.000Z,completeDate=2020-03-23T21:08:29.391Z,sequence=436]
        field_converters: Dict[str, Callable] = {
            "id": int,
            "rapidViewId": int,
            "sequence": int,
            "startDate": parse_date,
            "endDate": parse_date,
            "completeDate": parse_date,
        }
        field_finder = re.compile(r"[^\[]+\[(?P<fields>[^\]]*)\]")

        field_data_match = field_finder.match(representation)

        if not field_data_match:
            return None

        raw_field_data = field_data_match.groupdict()["fields"]
        gathered: Dict[str, Optional[Any]] = {}

        if raw_field_data:
            for field_str in raw_field_data.split(","):
                k, v = field_str.split("=")

                if v == "<null>":
                    gathered[k] = None
                    continue

                gathered[k] = field_converters.get(k, lambda x: x)(v)

        return SprintInfo(
            id=cast(int, gathered.get("id", -1)),
            name=cast(str, gathered.get("name", "")),
            state=cast(str, gathered.get("state", "")),
            goal=cast(str, gathered.get("goal", "")),
            rapidViewId=cast(Optional[int], gathered.get("rapidViewId")),
            sequence=cast(Optional[int], gathered.get("sequence")),
            startDate=cast(Optional[datetime.datetime], gathered.get("startDate")),
            endDate=cast(Optional[datetime.datetime], gathered.get("endDate")),
            completeDate=cast(
                Optional[datetime.datetime], gathered.get("completeDate")
            ),
        )

    def __call__(self, sprint_blob: Optional[str]) -> Optional[SprintInfo]:  # type: ignore[override]
        if sprint_blob is None:
            return None

        return self.get_sprint_details(sprint_blob)



import argparse
import getpass

import keyring

from ..constants import APP_NAME
from ..exceptions import UserError
from ..plugin import BaseCommand


class Command(BaseCommand):
    @classmethod
    def get_help(cls) -> str:
        return "Stores the password for a given user account in your system keychain."

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("username")

    def handle(self):
        password = getpass.getpass("Password: ")

        if not password:
            raise UserError("Password required")

        instance_url = self.config.get("instance_url") or self.options.instance_url
        keyring.set_password(APP_NAME, instance_url + self.options.username, password)



import argparse
import logging
import sys
from typing import Any
from typing import Dict
from typing import Iterable

from rich.console import Console
from rich.logging import RichHandler
from safdie import SafdieRunner

from .constants import COMMAND_ENTRYPOINT
from .exceptions import UserError
from .plugin import BaseCommand
from .utils import get_config
from .utils import get_default_config_path


class Runner(SafdieRunner):
    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        default_config_path = get_default_config_path()
        parser.add_argument(
            "--config",
            "-c",
            default=default_config_path,
            type=str,
            help=(
                "Path to configuration file to load; by default configuration "
                f"will be loaded from {default_config_path}."
            ),
        )
        parser.add_argument(
            "--instance-name",
            "-n",
            default="default",
            help=(
                "Named instance to connect to.  You can define new instances by "
                'running `jira-select configure --instance-name="my-instance"` '
                "to set the instance URL, username, and password for later use in "
                'other commands via providing the `--instance-name="my-instance" '
                "command-line argument.  By default the 'default' instance name "
                "will be used."
            ),
        )
        parser.add_argument(
            "--instance-url",
            "-i",
            default=None,
            type=str,
            help="URL of the Jira instance to connect to (if you cannot use 'configure')",
        )
        parser.add_argument(
            "--username",
            "-u",
            default=None,
            type=str,
            help="Username to use for connecting to Jira (if you cannot use 'configure')",
        )
        parser.add_argument(
            "--password",
            "-p",
            default=None,
            type=str,
            help=(
                "Password to use for connecting to Jira.  NOT RECOMMENDED: use the "
                "'store-password' or 'configure' command instead."
            ),
        )
        parser.add_argument(
            "--certificate",
            help=(
                "Path to a certificate to self-signed certificate use for "
                "connecting to your instance."
            ),
        )
        parser.add_argument(
            "--disable-certificate-verification",
            action="store_true",
            default=False,
            help="Do not verify server certificate.  Generally not recommended.",
        )
        parser.add_argument(
            "--debugger",
            action="store_true",
            help=(
                "Wait for debugger connection before processing command (requires debugpy)."
            ),
        )
        parser.add_argument(
            "--log-level",
            help=(
                "Print logging messages of the specified level and above "
                "to the console."
            ),
            default="INFO",
        )
        return super().add_arguments(parser)

    def handle(
        self,
        args: argparse.Namespace,
        init_args: Iterable[Any],
        init_kwargs: Dict[str, Any],
        handle_args: Iterable[Any],
        handle_kwargs: Dict[str, Any],
    ) -> Any:
        console = Console()
        logging.basicConfig(
            level=logging.getLevelName(args.log_level),
            format="%(message)s",
            datefmt="[%X]",
            handlers=[RichHandler(rich_tracebacks=True, console=Console(stderr=True))],
        )

        if args.debugger:
            import debugpy

            debugger_host = "0.0.0.0"
            debugger_port = 5678

            debugpy.listen((debugger_host, debugger_port))

            console.print(
                f"[magenta]Waiting for debugger connection on {debugger_host}:{debugger_port}[/magenta]..."
            )

            # Pause the program until a remote debugger is attached
            debugpy.wait_for_client()

        config = get_config(path=args.config)

        init_kwargs["config"] = config
        return super().handle(args, init_args, init_kwargs, handle_args, handle_kwargs)


def main():
    console = Console()

    try:
        Runner(COMMAND_ENTRYPOINT, cmd_class=BaseCommand).run()
    except UserError as e:
        console.print(f"[red]{e}[/red]")
        sys.exit(1)
    except Exception:
        console.print_exception()
        sys.exit(1)



import errno
import time
from typing import Any
from typing import Optional

from diskcache.core import EVICTION_POLICY
from diskcache.core import Cache


class MinimumRecencyCache(Cache):
    def get(
        self,
        key,
        default=None,
        read=False,
        expire_time=False,
        tag=False,
        retry=False,
        min_recency=2**32,
    ) -> Optional[Any]:
        """Retrieve value from cache. If `key` is missing, return `default`.

        .. note::

           This is lifted directly from the `diskcache.core.Cache.get` method
           in the superclass; just adding logic around a minimum recency
           restriction

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).
        :param key: key for item
        :param default: value to return if key is missing (default None)
        :param bool read: if True, return file handle to value
            (default False)
        :param bool expire_time: if True, return expire_time in tuple
            (default False)
        :param bool tag: if True, return tag in tuple (default False)
        :param bool retry: retry if database timeout occurs (default False)
        :param int min_recency: return the value only if it is fewer than
            this number of seconds old.
        :return: value for item or default if key not found
        :raises Timeout: if database timeout occurs
        """
        db_key, raw = self._disk.put(key)
        update_column = EVICTION_POLICY[self.eviction_policy]["get"]
        select = (
            "SELECT rowid, expire_time, tag, mode, filename, value"
            " FROM Cache WHERE key = ? AND raw = ?"
            " AND (expire_time IS NULL OR expire_time > ?)"
            " AND (store_time >= ?) "
        )

        if expire_time and tag:
            default = (default, None, None)
        elif expire_time or tag:
            default = (default, None)

        if not self.statistics and update_column is None:
            # Fast path, no transaction necessary.

            rows = self._sql(
                select, (db_key, raw, time.time(), time.time() - min_recency)
            ).fetchall()

            if not rows:
                return default

            ((rowid, db_expire_time, db_tag, mode, filename, db_value),) = rows

            try:
                value = self._disk.fetch(mode, filename, db_value, read)
            except OSError:
                # Key was deleted before we could retrieve result.
                return default

        else:  # Slow path, transaction required.
            cache_hit = 'UPDATE Settings SET value = value + 1 WHERE key = "hits"'
            cache_miss = 'UPDATE Settings SET value = value + 1 WHERE key = "misses"'

            with self._transact(retry) as (sql, _):
                rows = sql(
                    select, (db_key, raw, time.time(), time.time() - min_recency)
                ).fetchall()

                if not rows:
                    if self.statistics:
                        sql(cache_miss)
                    return default

                ((rowid, db_expire_time, db_tag, mode, filename, db_value),) = rows

                try:
                    value = self._disk.fetch(mode, filename, db_value, read)
                except OSError as error:
                    if error.errno == errno.ENOENT:
                        # Key was deleted before we could retrieve result.
                        if self.statistics:
                            sql(cache_miss)
                        return default
                    else:
                        raise

                if self.statistics:
                    sql(cache_hit)

                now = time.time()
                update = "UPDATE Cache SET %s WHERE rowid = ?"

                if update_column is not None:
                    sql(update % update_column.format(now=now), (rowid,))

        if expire_time and tag:
            return (value, db_expire_time, db_tag)
        elif expire_time:
            return (value, db_expire_time)
        elif tag:
            return (value, db_tag)
        else:
            return value



import csv
from io import TextIOWrapper
from typing import Any
from typing import Dict
from typing import List

from ..plugin import BaseFormatter


class Formatter(BaseFormatter):
    delimiter = ","

    @classmethod
    def get_file_extension(cls) -> str:
        return "csv"

    def _generate_fieldnames(self) -> List[str]:
        fields = []

        for field in self.executor.query.select:
            fields.append(field.column)

        return fields

    def open(self):
        super().open()
        self._wrapped_stream = TextIOWrapper(
            self.stream, encoding="utf-8", write_through=True, newline=""
        )
        self.out = csv.DictWriter(
            self._wrapped_stream,
            fieldnames=self._generate_fieldnames(),
            delimiter=self.delimiter,
        )
        self.out.writeheader()

    def close(self):
        self._wrapped_stream.detach()

    def writerow(self, row: Dict[str, Any]):
        self.out.writerow(row)



import argparse
import importlib.util
import os
import sys
import tempfile

from ..exceptions import UserError
from ..plugin import BaseCommand
from ..utils import get_custom_function_dir


class Command(BaseCommand):
    @classmethod
    def get_help(cls) -> str:
        path = get_custom_function_dir()
        return (
            "Installs a user script (i.e. a python script having functions "
            "decorated with `jira_select.plugin.register_function`) into "
            f"your user configuration directory at {path}."
        )

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "python_script",
            type=argparse.FileType("r"),
            help=("Script to install into your user scripts directory."),
        )
        parser.add_argument(
            "--overwrite",
            action="store_true",
            help=(
                "By default, if the specified filename matches an existing "
                "user script, an exception will be raised. If you would like "
                "to overwrite an existing file, though, you can use this "
                "argument."
            ),
        )
        parser.add_argument(
            "--name",
            default=None,
            help=(
                "Override the name of your generated script; if not named "
                "ending with `.py`, `.py` will be appended automatically."
            ),
        )
        return super().add_arguments(parser)

    def handle(self, *args, **kwargs) -> None:
        name = self.options.name
        if name is None and self.options.python_script is not sys.stdin:
            name = self.options.python_script.name
        if name is None:
            raise UserError(
                "When creating a script file from stdin, you must "
                "provide the --name command-line argument."
            )
        if not name.endswith(".py"):
            name += ".py"

        full_path = os.path.join(get_custom_function_dir(), name)
        if not self.options.overwrite and os.path.exists(full_path):
            raise UserError(
                f"Path {full_path} already exists; use --overwrite to overwrite."
            )

        with tempfile.NamedTemporaryFile(mode="w+", suffix=".py") as tmpf:
            tmpf.write(self.options.python_script.read())
            tmpf.flush()
            tmpf.seek(0)

            try:
                spec = importlib.util.spec_from_file_location(
                    "temporary_user_script", tmpf.name
                )
                assert spec
                user_module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(user_module)  # type: ignore
            except Exception as e:
                raise UserError(f"Could not install user script: {e}")

            tmpf.seek(0)
            with open(full_path, "w") as outf:
                outf.write(tmpf.read())



from __future__ import annotations

import uuid
from unittest.mock import Mock

import pytest
from dotmap import DotMap
from jira import Issue
from jira.resources import Resource

from jira_select.exceptions import ExpressionParameterMissing
from jira_select.query import Executor
from jira_select.types import QueryDefinition

from .base import JiraSelectTestCase


class JiraList(list):
    pass


class MyResource(Resource):
    def __init__(self, raw):
        super().__init__(None, None, None)
        self.raw = raw


class NonResource:
    def __str__(self):
        return "<NonResource>"


class TestQuery(JiraSelectTestCase):
    def setUp(self):
        super().setUp()

        self.JIRA_ISSUES = [
            {
                "key": "ALPHA-1",
                "fields": {
                    "issuetype": "Issue",
                    "summary": "My Ticket",
                    "project": "ALPHA",
                    "story_points": 1,
                    "customfield10010": 50,
                    "customfield10011": "Ivanovna",
                    "customfield10012": MyResource({"ok": "yes"}),
                    "customfield10013": NonResource(),
                    "worklogs": DotMap(
                        {"total": 1, "worklogs": [{"timespentSeconds": 60}]}
                    ),
                    "transactions": DotMap(
                        {"byCurrency": {"usd": 100, "cad": 105, "cop": 200}}
                    ),
                },
            },
            {
                "key": "ALPHA-3",
                "fields": {
                    "issuetype": "Bug",
                    "summary": "Another Ticket",
                    "project": "ALPHA",
                    "story_points": 10,
                    "customfield10010": 55,
                    "customfield10011": "Jackson",
                    "customfield10012": MyResource({"ok": "no"}),
                    "customfield10013": NonResource(),
                    "worklogs": DotMap(
                        {
                            "total": 2,
                            "worklogs": [
                                {"timespentSeconds": 30},
                                {"timespentSeconds": 15},
                            ],
                        }
                    ),
                    "transactions": DotMap({"byCurrency": {"cad": 124, "cop": 843}}),
                },
            },
            {
                "key": "ALPHA-2",
                "fields": {
                    "issuetype": "Issue",
                    "summary": "My Ticket #2",
                    "project": "ALPHA",
                    "story_points": 1,
                    "customfield10010": 56,
                    "customfield10011": "Chartreuse",
                    "customfield10012": MyResource({"ok": "maybe"}),
                    "customfield10013": NonResource(),
                    "transactions": DotMap({"byCurrency": {"usd": 10, "rur": 33}}),
                },
            },
        ]
        issues = JiraList([Issue(None, None, issue) for issue in self.JIRA_ISSUES])
        issues.total = len(self.JIRA_ISSUES)

        self.mock_jira = Mock(
            search_issues=Mock(return_value=issues), fields=Mock(return_value=[])
        )

    def test_simple(self):
        query = QueryDefinition.parse_obj(
            {
                "select": ["key"],
                "from": "issues",
            }
        )

        query = Executor(self.mock_jira, query)

        actual_results = list(query)
        expected_results = [
            {
                "key": "ALPHA-1",
            },
            {
                "key": "ALPHA-3",
            },
            {
                "key": "ALPHA-2",
            },
        ]

        assert expected_results == actual_results

    def test_sort_by(self):
        query = QueryDefinition.parse_obj(
            {
                "select": ["key"],
                "from": "issues",
                "sort_by": ["story_points desc", "key"],
            }
        )

        query = Executor(self.mock_jira, query)

        actual_results = list(query)
        expected_results = [
            {
                "key": "ALPHA-3",
            },
            {
                "key": "ALPHA-1",
            },
            {
                "key": "ALPHA-2",
            },
        ]

        assert expected_results == actual_results

    def test_filter(self):
        query = QueryDefinition.parse_obj(
            {
                "select": ["key"],
                "from": "issues",
                "filter": ["summary != 'My Ticket #2'"],
            }
        )

        query = Executor(self.mock_jira, query)

        actual_results = list(query)
        expected_results = [
            {
                "key": "ALPHA-1",
            },
            {
                "key": "ALPHA-3",
            },
        ]

        assert expected_results == actual_results

    def test_group_by_aggregation(self):
        query = QueryDefinition.parse_obj(
            {
                "select": ["issuetype", "len(key)"],
                "from": "issues",
                "group_by": ["issuetype"],
                "sort_by": ["len(key)"],
            }
        )

        query = Executor(self.mock_jira, query)

        actual_results = list(query)
        expected_results = [
            {
                "issuetype": "Bug",
                "len(key)": 1,
            },
            {
                "issuetype": "Issue",
                "len(key)": 2,
            },
        ]

        assert expected_results == actual_results

    def test_cap(self):
        query = QueryDefinition.parse_obj(
            {
                "select": ["key"],
                "from": "issues",
                "cap": 1,
            }
        )

        query = Executor(self.mock_jira, query)

        actual_results = list(query)
        assert len(actual_results) == 1

    def test_simple_wprogress(self):
        query = QueryDefinition.parse_obj(
            {
                "select": ["key"],
                "from": "issues",
            }
        )

        query = Executor(self.mock_jira, query, True)

        actual_results = list(query)
        expected_results = [
            {
                "key": "ALPHA-1",
            },
            {
                "key": "ALPHA-3",
            },
            {
                "key": "ALPHA-2",
            },
        ]

        assert expected_results == actual_results

    def test_field_name_map(self):
        arbitrary_query = QueryDefinition.parse_obj(
            {
                "select": ["key"],
                "from": "issues",
            }
        )
        self.mock_jira.fields = Mock(
            return_value=[
                {"name": "Story Points", "id": "customfield10010"},
                {"name": "Sprint", "id": "customfield10011"},
            ]
        )
        query = Executor(self.mock_jira, arbitrary_query, True)

        expected_result = {
            "Story Points": "customfield10010",
            "Sprint": "customfield10011",
            "params": DotMap(),
        }
        actual_result = query.field_name_map

        assert expected_result == actual_result

    def test_interpolated_value_nonspecial(self):
        arbitrary_query = QueryDefinition.parse_obj(
            {
                "select": ['{customfield10011} as "mn"'],
                "from": "issues",
            }
        )
        self.mock_jira.fields = Mock(
            return_value=[
                {"name": "Maiden Name", "id": "customfield10011"},
            ]
        )

        query = Executor(self.mock_jira, arbitrary_query, True)

        actual_results = list(query)
        expected_results = [
            {"mn": "Ivanovna"},
            {"mn": "Jackson"},
            {"mn": "Chartreuse"},
        ]

        assert expected_results == actual_results

    def test_interpolated_value_nonstring(self):
        arbitrary_query = QueryDefinition.parse_obj(
            {
                "select": ['{Jellybean Guess} as "jb"'],
                "from": "issues",
            }
        )
        self.mock_jira.fields = Mock(
            return_value=[
                {"name": "Jellybean Guess", "id": "customfield10010"},
            ]
        )

        query = Executor(self.mock_jira, arbitrary_query, True)

        actual_results = list(query)
        expected_results = [
            {"jb": 50},
            {"jb": 55},
            {"jb": 56},
        ]

        assert expected_results == actual_results

    def test_interpolated_value_string(self):
        arbitrary_query = QueryDefinition.parse_obj(
            {
                "select": ['{Maiden Name} as "mn"'],
                "from": "issues",
            }
        )
        self.mock_jira.fields = Mock(
            return_value=[
                {"name": "Maiden Name", "id": "customfield10011"},
            ]
        )

        query = Executor(self.mock_jira, arbitrary_query, True)

        actual_results = list(query)
        expected_results = [
            {"mn": "Ivanovna"},
            {"mn": "Jackson"},
            {"mn": "Chartreuse"},
        ]

        assert expected_results == actual_results

    def test_interpolated_value_resource(self):
        arbitrary_query = QueryDefinition.parse_obj(
            {
                "select": ['{Result Object}.ok as "ro"'],
                "from": "issues",
            }
        )
        self.mock_jira.fields = Mock(
            return_value=[
                {"name": "Result Object", "id": "customfield10012"},
            ]
        )

        query = Executor(self.mock_jira, arbitrary_query, True)

        actual_results = list(query)
        expected_results = [
            {"ro": "yes"},
            {"ro": "no"},
            {"ro": "maybe"},
        ]

        assert expected_results == actual_results

    def test_interpolated_value_nonresource(self):
        arbitrary_query = QueryDefinition.parse_obj(
            {
                "select": ['{Beep Boop} as "bb"'],
                "from": "issues",
            }
        )
        self.mock_jira.fields = Mock(
            return_value=[
                {"name": "Beep Boop", "id": "customfield10013"},
            ]
        )

        query = Executor(self.mock_jira, arbitrary_query, True)

        actual_results = list(query)
        for result in actual_results:
            assert isinstance(result["bb"], NonResource)

    def test_autoextract_sum(self):
        query = QueryDefinition.parse_obj(
            {
                "select": ['sum(transactions.byCurrency.usd) as "value"'],
                "from": "issues",
                "group_by": ["True"],
            }
        )
        query = Executor(self.mock_jira, query)

        actual_results = list(query)[0]["value"]
        expected_results = 110

        assert expected_results == actual_results

    def test_autoextract_internal_array(self):
        query = QueryDefinition.parse_obj(
            {
                "select": [
                    'sum(extract(flatten_list(worklogs.worklogs), "timespentSeconds")) as "value"'
                ],
                "from": "issues",
                "group_by": ["True"],
            }
        )
        query = Executor(self.mock_jira, query)

        actual_results = list(query)[0]["value"]
        expected_results = 105

        assert expected_results == actual_results

    def test_param_expression_specified(self):
        arbitrary_value = str(uuid.uuid4())

        query = QueryDefinition.parse_obj(
            {
                "select": ["'{params.ok}' as \"value\""],
                "from": "issues",
                "cap": 1,
            }
        )

        query = Executor(self.mock_jira, query, parameters={"ok": arbitrary_value})
        results = list(query)

        assert results[0]["value"] == arbitrary_value

    def test_param_expression_unspecified_select(self):
        query = QueryDefinition.parse_obj(
            {
                "select": ["'{params.ok}'"],
                "from": "issues",
                "cap": 1,
            }
        )

        query = Executor(self.mock_jira, query)

        with pytest.raises(ExpressionParameterMissing):
            list(query)

    def test_param_expression_unspecified_where(self):
        query = QueryDefinition.parse_obj(
            {
                "select": ["True"],
                "from": "issues",
                "where": [
                    "'{params.ok}'",
                ],
                "cap": 1,
            }
        )

        query = Executor(self.mock_jira, query)

        with pytest.raises(ExpressionParameterMissing):
            list(query)

    def test_param_jql_specified(self):
        arbitrary_value = str(uuid.uuid4())

        query = QueryDefinition.parse_obj(
            {
                "select": ["'{params.ok}' as \"value\""],
                "from": "issues",
                "where": [
                    '"{params.ok}"',
                ],
                "cap": 1,
            }
        )

        query = list(
            Executor(self.mock_jira, query, parameters={"ok": arbitrary_value})
        )

        args, _ = self.mock_jira.search_issues.call_args

        assert arbitrary_value in args[0]



from typing import Optional

from .sprint_details import Function as SprintDetailsFunction


class Function(SprintDetailsFunction):
    """Returns the name of the sprint matching the provided ID."""

    def __call__(self, sprint_id: Optional[str]) -> Optional[str]:  # type: ignore[override]
        if sprint_id is None:
            return None

        info = self.get_sprint_details(sprint_id)
        if info is None:
            return None

        return info.name



import json
from io import TextIOWrapper
from typing import Any
from typing import Dict
from typing import List

from ..plugin import BaseFormatter
from ..utils import JiraSelectJsonEncoder


class Formatter(BaseFormatter):
    _rows: List[Dict] = []

    @classmethod
    def get_file_extension(cls) -> str:
        return "json"

    def open(self):
        super().open()
        self._rows = []

    def close(self):
        wrapped = TextIOWrapper(self.stream, encoding="utf-8", write_through=True)
        json.dump(self._rows, wrapped, cls=JiraSelectJsonEncoder, indent=4)
        wrapped.detach()
        super().close()

    def writerow(self, row: Dict[str, Any]):
        self._rows.append(row)



import datetime
from dataclasses import dataclass
from typing import Any
from typing import Iterator
from typing import List
from typing import Optional

from dateutil.parser import parse as parse_date

from ..exceptions import UserError
from ..plugin import BaseFunction


@dataclass
class ChangelogEntry:
    author: str = ""
    created: datetime.datetime = datetime.datetime.utcnow()
    id: int = -1
    field: str = ""
    fieldtype: str = ""
    fromValue: Optional[Any] = None
    fromString: Optional[str] = None
    toValue: Optional[Any] = None
    toString: Optional[str] = None


def flatten_changelog(changelog: Any) -> Iterator[ChangelogEntry]:
    for entry in changelog.histories:
        author = str(entry.author)
        created = parse_date(entry.created)
        id = int(entry.id)

        for item in entry.items:
            yield ChangelogEntry(
                author=author,
                created=created,
                id=id,
                field=item.field,
                fieldtype=item.fieldtype,
                fromValue=getattr(item, "from"),  # noqa
                fromString=item.fromString,
                toValue=getattr(item, "to"),  # noqa
                toString=item.toString,
            )


class Function(BaseFunction):
    """Returns a flattened list of changelog entries.


    Returns ``jira_select.functions.flatten_changelog.ChangelogEntry``
    entries for every changelog entry.
    """

    def __call__(self, changelog: Any) -> List[ChangelogEntry]:  # type: ignore[override]
        if changelog is None or not hasattr(changelog, "histories"):
            raise UserError(
                "No changelog was provided; did you use the ``expand`` "
                "option of ``changelog`` in your query?"
            )

        return list(flatten_changelog(changelog))



from __future__ import annotations

import copy
import datetime
from typing import Any
from typing import Iterator

from dateutil.parser import parse as parse_datetime
from dotmap import DotMap
from jira.resources import Issue
from pytz import UTC

from jira_select.plugin import BaseFunction

from ..exceptions import QueryError
from .flatten_changelog import flatten_changelog


class IssueSnapshotContainer(dict):
    _name_map: dict[str, Any] = {}

    def __init__(self, field_dict: dict[str, Any], name_map: dict[str, str]) -> None:
        self._name_map = name_map

        super().__init__(field_dict)

    def __getitem__(self, item: str) -> Any:
        if item not in self and item in self._name_map:
            return super().__getitem__(self._name_map[item])

        return super().__getitem__(item)

    def __setitem__(self, item: str, value: Any) -> None:
        if item not in self and item in self._name_map:
            return super().__setitem__(self._name_map[item], value)

        return super().__setitem__(item, value)


def snapshot_iterator(issue: Issue, field_name_map: dict[str, str]) -> Iterator[DotMap]:
    non_snapshottable = [
        "changelog",
        "components",
        "comment",
        "expand",
        "parent",
        "raw",
        "self",
        "subtasks",
        "timetracking",
        "updated",
        "votes",
        "watches",
        "worklog",
    ]
    snapshot: dict[str, Any] = IssueSnapshotContainer(
        {
            k: (str(v) if v is not None else None)
            for k, v in issue.as_dict().items()
            if not callable(v) and k not in non_snapshottable
        },
        field_name_map,
    )
    snapshot_validity_end = datetime.datetime.utcnow().replace(tzinfo=UTC)

    try:
        flattened_changelog = sorted(
            flatten_changelog(issue.changelog),
            key=lambda row: row.created if row.created else datetime.date(1, 1, 1),
            reverse=True,
        )
    except AttributeError as exc:
        raise QueryError(
            "'expand' option of 'changelog' is required for snapshot iteration; "
        ) from exc

    for entry in flattened_changelog:
        if entry.field in non_snapshottable:
            continue

        snapshot.update(
            {
                "validity_start": entry.created,
                "validity_end": snapshot_validity_end,
            }
        )

        yield DotMap(copy.deepcopy(snapshot))

        snapshot_validity_end = entry.created
        snapshot[entry.field] = (
            str(entry.fromString) if entry.fromString is not None else None
        )

    yield DotMap(copy.deepcopy(snapshot))


class Function(BaseFunction):
    """Returns an IssueSnapshot representing the Jira Issue's state on a particular date."""

    _field_name_map = None

    def __call__(  # type: ignore[override]
        self, issue: Issue, date: datetime.datetime
    ) -> DotMap:
        last_snapshot: DotMap = DotMap({})

        if date > parse_datetime(issue.created):
            for snapshot in snapshot_iterator(
                issue, self.executor.field_name_map if self.executor else {}
            ):
                if parse_datetime(snapshot.created) < date:
                    return last_snapshot

                last_snapshot = snapshot

        return last_snapshot



from typing import Any
from typing import Dict
from typing import Iterable

from QueryableList import QueryableListMixed

from ..plugin import BaseFunction


class Function(BaseFunction):
    """Returns entries in an iterable matching a provided django-style query.

    Returned rows are required to match at least one of the provided query
    parameters.

    """

    def __call__(  # type: ignore[override]
        self, iterable: Iterable[Any], **query: Dict[str, Any]
    ) -> Iterable[Any]:
        queryable = QueryableListMixed(iterable)

        return queryable.filterOr(**query)



import datetime

from dateutil.tz import tzlocal

from ..plugin import BaseFunction


class Function(BaseFunction):
    """Return 'now' as a timezone-aware datetime object."""

    def __call__(self, **replacements) -> datetime.datetime:  # type: ignore[override]
        localized = (
            datetime.datetime.now().replace(tzinfo=tzlocal()).replace(**replacements)
        )

        return localized



#!/usr/bin/env python

import io
import os
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup

requirements_path = os.path.join(
    os.path.dirname(__file__),
    "requirements.txt",
)
requirements = []
with open(requirements_path, "r") as in_:
    requirements = [
        req.strip()
        for req in in_.readlines()
        if not req.startswith("-") and not req.startswith("#")
    ]


def read(*names, **kwargs):
    with io.open(
        join(dirname(__file__), *names), encoding=kwargs.get("encoding", "utf8")
    ) as fh:
        return fh.read()


setup(
    name="jira-select",
    version="3.1.0",
    license="MIT",
    description=(
        "Easily run complex SQL-like queries far beyond what "
        "Jira's standard JQL query language can provide."
    ),
    long_description_content_type="text/markdown",
    long_description=read("readme.md"),
    author="Adam Coddington",
    author_email="me@adamcoddington.net",
    url="https://github.com/coddingtonbear/jira-select",
    packages=find_packages(),
    py_modules=[splitext(basename(path))[0] for path in glob("jira-select/*.py")],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Programming Language :: Python :: 3",
        "Topic :: Utilities",
    ],
    project_urls={
        "Issue Tracker": "https://github.com/coddingtonbear/jira-select/issues",
    },
    keywords=[
        "jira",
        "csv",
        "sql",
    ],
    python_requires=">=3.6",
    install_requires=requirements,
    extras_require={},
    entry_points={
        "console_scripts": ["jira-select = jira_select.cmdline:main"],
        "jira_select.commands": [
            "configure = jira_select.commands.configure:Command",
            "store-password = jira_select.commands.store_password:Command",
            "build-query = jira_select.commands.build_query:Command",
            "run-query = jira_select.commands.run:Command",
            "schema = jira_select.commands.schema:Command",
            "functions = jira_select.commands.functions:Command",
            "shell = jira_select.commands.shell:Command",
            "run-script = jira_select.commands.run_script:Command",
            "show-instances = jira_select.commands.show_instances:Command",
            "setup-instance = jira_select.commands.setup_instance:Command",
            "remove-instance = jira_select.commands.remove_instance:Command",
            "install-user-script = jira_select.commands.install_user_script:Command",
        ],
        "jira_select.formatters": [
            "csv = jira_select.formatters.csv:Formatter",
            "tsv = jira_select.formatters.tsv:Formatter",
            "html = jira_select.formatters.html:Formatter",
            "json = jira_select.formatters.json:Formatter",
            "table = jira_select.formatters.table:Formatter",
            "raw = jira_select.formatters.raw:Formatter",
        ],
        "jira_select.functions": [
            "sprint_details = jira_select.functions.sprint_details:Function",
            "sprint_name = jira_select.functions.sprint_name:Function",
            "get_sprint_by_id = jira_select.functions.get_sprint_by_id:Function",
            "get_sprint_by_name = jira_select.functions.get_sprint_by_name:Function",
            "field_by_name = jira_select.functions.field_by_name:Function",
            "estimate_to_days = jira_select.functions.estimate_to_days:Function",
            "extract = jira_select.functions.extract:Function",
            "flatten_list = jira_select.functions.flatten_list:Function",
            "flatten_changelog = jira_select.functions.flatten_changelog:Function",
            "simple_filter = jira_select.functions.simple_filter:Function",
            "simple_filter_any = jira_select.functions.simple_filter_any:Function",
            "parse_datetime = jira_select.functions.parse_datetime:Function",
            "parse_date = jira_select.functions.parse_date:Function",
            "json_dumps = jira_select.functions.json_dumps:Function",
            "get_issue = jira_select.functions.get_issue:Function",
            "workdays_in_state = jira_select.functions.workdays_in_state:Function",
            "get_issue_snapshot_on_date = jira_select.functions.get_issue_snapshot_on_date:Function",
            "interval_business_hours = jira_select.functions.interval_business_hours:Function",
            "interval_matching = jira_select.functions.interval_matching:Function",
            "interval_size = jira_select.functions.interval_size:Function",
            "subquery = jira_select.functions.subquery:Function",
            "union = jira_select.functions.union:Function",
            "now = jira_select.functions.now:Function",
        ],
        "jira_select.sources": [
            "issues = jira_select.sources.issues:Source",
            "boards = jira_select.sources.boards:Source",
            "sprints = jira_select.sources.sprints:Source",
        ],
    },
)



import argparse
import importlib.machinery

from ..plugin import BaseCommand


class Command(BaseCommand):
    @classmethod
    def get_help(cls) -> str:
        return (
            "Runs the 'main(**kwargs)' function in the specified file; "
            "see documentation for details."
        )

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "file_path",
            type=str,
            help=(
                "Path to the file containg the `main(**kwargs)` function" "to execute."
            ),
        )
        parser.add_argument("command_args", nargs=argparse.REMAINDER)

    def handle(self):
        run_script_module = importlib.machinery.SourceFileLoader(
            "run_script", self.options.file_path
        ).load_module()
        run_script_module.main(args=self.options.command_args, cmd=self)



from typing import Dict
from typing import Iterator
from typing import List

from jira.resources import Board

from ..exceptions import QueryError
from ..plugin import BaseSource
from ..types import SchemaRow


class Source(BaseSource):
    SCHEMA: List[SchemaRow] = [
        SchemaRow.parse_obj({"id": "id", "type": "int"}),
        SchemaRow.parse_obj({"id": "name", "type": "str"}),
        SchemaRow.parse_obj({"id": "type", "type": "str"}),
    ]

    def __iter__(self) -> Iterator[Dict]:
        start_at = 0
        max_results = 2**32
        result_limit = self.query.limit or 2**32

        if self.query.order_by:
            raise QueryError(
                "Board query 'order_by' expressions are not supported. "
                "Use 'sort_by' instead."
            )
        if self.query.expand:
            raise QueryError("Board query 'expand' expressions are not supported.")

        where = self.query.where or {}
        if where and not isinstance(where, dict):
            raise QueryError(
                "Board query 'where' expressions should be a dictionary "
                "having any of the following keys: 'type' or 'name'"
            )

        param_type = where.pop("type", None)
        param_name = where.pop("name", None)

        if where:
            raise QueryError(f"Unexpected 'where' parameters: {where}.")

        self.update_progress(completed=0, total=1, visible=True)
        while start_at < min(max_results, result_limit):
            results = self.jira.boards(
                startAt=start_at,
                maxResults=min(result_limit, 100),
                type=param_type,
                name=param_name,
            )

            max_results = results.total
            count = min([results.total, result_limit])
            self.update_count(count)

            for result in results:
                self.update_progress(advance=1, total=count, visible=True)

                yield result.raw

                start_at += 1

                # Return early if our result limit has been reached
                if start_at >= result_limit:
                    break

    def rehydrate(self, value: Dict) -> Board:
        return Board(
            {"agile_rest_path": self.jira._options["agile_rest_path"]}, None, value
        )
